from .fx_return import FXReturn
from .group import Group
from .input import Input
from .fx_return import FXReturn
from .fx_send import FXSend
from .lr import LR
from .matrix import Matrix
from ._aux import Aux


from sq_midi.parameter_wrapper import ParameterWrapper

class ParameterHandler:
    """ Handle setting / getting parameters """
    def __init__(self, wrapper: ParameterWrapper, parameter_name):
        self._mixer = wrapper.__getattribute__("mixer")
        self._list = wrapper.__getattribute__("channel").info[
            wrapper.__getattribute__("parameter_category")
        ][parameter_name]
        #print(self._list)
        self._data_type = wrapper.__getattribute__("parameter_type")
        self._data_range = wrapper.__getattribute__("parameter_range")
        self._data_map = wrapper.__getattribute__("parameter_map")

    def __getitem__(self, key: int):
        """ Get a parameter """
        if key not in range(1, len(self._list) + 1):
            raise IndexError(f"Parameter index out of range [1, {len(self._list)}]")
        #print(self._list[key-1])
        return self._mixer.get_param(self._list[key-1])

    def __setitem__(self, key, value):
        """ Set a parameter
            1-indexed, e.g. aux[1] => Aux1
        """
        if type(key) is not int:
            raise TypeError("Parameter key must be an integer")
        if key not in range(1, len(self._list) + 1):
            raise IndexError(f"Parameter index out of range [1, {len(self._list)}]")
        if (
            type(value) is not self._data_type
            and not (
                self._data_type is bool
                and isinstance(value, int)
                and value in [0, 1]
            )
        ):
            raise ValueError(f"Value must be of type {str(self._data_type)}")
        if value not in self._data_range:
            raise ValueError("Value must be in range " + str(self._data_range[0]) + " to " + str(self._data_range[-1]))
        value = self._data_map(value)
        coarse = value >> 7
        fine = value & 0x7F
        #print("Coarse: " + str(coarse), "Fine: " + str(fine))
        self._mixer.set_param(self._list[key - 1], coarse, fine)


from gc import freeze


class ChannelWrapper:

    __freeze = False

    def __init__(self, channel_list: list):
        self.__channel_list = channel_list
        self.__prefix = self.__channel_list[0].CHANNEL_PREFIX
        self.__type = self.__channel_list[0].CHANNEL_TYPE
        __freeze = True

    def __str__(self):
        return "Mixer " + self.__type + " [total " + str(len(self.__channel_list)) + "]"

    def __dir__(self):
        return [f"{self.__prefix}{i}" for i in range(1, len(self.__channel_list)+1)]

    def __getitem__(self, key):
        """ Get channel by number """
        if key not in range(1, len(self.__channel_list) + 1):
            raise IndexError(f"Parameter index must be in range [1, {len(self.__channel_list)}]")
        return self.__channel_list[key - 1]

    def __getattr__(self, attr):
        """ Get channel by name """
        item = attr.capitalize()
        if item.startswith(self.__prefix.capitalize()) and item.replace(self.__prefix.capitalize(), '').isdigit():
            return self.__channel_list[int(item.replace(self.__prefix.capitalize(), '')) - 1]
        else:
            # Handle default
            return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if self.__freeze:
            raise TypeError("Channel lists are read-only")
        object.__setattr__(self, key, value)


from sq_midi.channel import Channel
from sq_midi.mixer import Mixer


class FXReturn(Channel):
    """ FX Return Mix Class """

    FILE_NAME = 'fx_returns'
    CHANNEL_TYPE = 'FX Return'
    CHANNEL_PREFIX = "FxRtn"
    NUMBER_OF_CHANNELS = 8

if __name__ == '__main__':
    _mixer = Mixer(debug=True)
    fx = FXReturn(_mixer, 8)

    print(fx.assignments.groups[1])



import json
import os

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))

def decode(data):
    if isinstance(data, dict):
        return {key: decode(val) for key, val in data.items()}
    elif isinstance(data, list):
        return [decode(val) for val in data]
    elif isinstance(data, str):
        try:
            return int(data, 16)
        except ValueError:
            return data
    else:
        return data

def load_data(file_name: str):
    with open(os.path.join(__location__, file_name + ".json")) as data:
        return decode(json.load(data))

if __name__ == '__main__':
    test = load_data("inputs")
    print(test[0])


from typing import Callable

class ParameterWrapper:
    """ Parameter wrapper class
        handles parameters
    """

    from sq_midi import Mixer
    from sq_midi.channel import Channel

    channel: Channel = None
    mixer: Mixer = None
    parameter_category: str = None
    parameter_type: object = None
    parameter_range: list = None
    parameter_map: Callable = None


    def __init__(
            self,
            mixer: Mixer,
            channel: Channel,
            parameter_category: str,
            parameter_type: object,
            parameter_range: list,
            parameter_map: Callable
        ):
        self.channel = channel
        self.mixer = mixer
        self.parameter_category = parameter_category
        self.parameter_type = parameter_type
        self.parameter_range = parameter_range
        self.parameter_map = parameter_map
        #print("Initializing Parameter Wrapper with " + parameter_category)

    def __getattr__(self, item):
        """ Handle parameter destination by name
            e.g. mixer.inputs[1].aux
                 mixer.inputs[1].mtx
        """
        #print("__getattr__ called " + item)
        if item not in self.channel.info[self.parameter_category]:
            raise ValueError(
                f"'{self.parameter_category}' parameter does not have destination type '{item}'"
                f"\nAvailable values: { ", ".join(self.channel.info[self.parameter_category].keys())}"
            )
        # Load handler here (avoiding circular dependency)
        from sq_midi.parameter_handler import ParameterHandler
        handler = ParameterHandler(self, item)
        # Return handler
        # If LR, then return directly
        return handler[1] if item == "lr" else handler

    def __setattr__(self, key, value):
        """ Used to set LR parameters (not indexed) """
        if key == "lr":
            from sq_midi.parameter_handler import ParameterHandler
            ParameterHandler(self, key)[1] = value
        else:
            if getattr(ParameterWrapper, key, False) is not False:
                super().__setattr__(key, value)
            else:
                raise ValueError(
                    f"'{self.parameter_category}' parameter does not have destination type '{key}'"
                    f"\nAvailable values: {", ".join(self.channel.info[self.parameter_category].keys())}"
                )


from sq_midi import Channel

class MuteGroup(Channel):
    """ Mute groups """

    CHANNEL_TYPE = "Mute Group"
    CHANNEL_PREFIX = "MGRP"
    NUMBER_OF_CHANNELS = 8

    def _load_info(self):
        return {
            "name": self.CHANNEL_PREFIX + str(self.number),
            "mute": ["04", "0" + str(self.number-1)],
        }

if __name__ == "__main__":
    from sq_midi import Mixer
    c = MuteGroup(Mixer(), 8)
    print(c.info)


from sq_midi.channel import Channel

class FXSend(Channel):
    """ FX Send Class """

    FILE_NAME = 'fx_sends'
    CHANNEL_TYPE = 'FX Send'
    CHANNEL_PREFIX = "FxSnd"
    CHANNEL_PARAMETERS = [
        "levels"
    ]
    NUMBER_OF_CHANNELS = 4

if __name__ == '__main__':
    from sq_midi import Mixer

    test = FXSend(Mixer(), 1)

    print(test)


import time
import mido
from mido import Message

from sq_midi.channel_wrapper import ChannelWrapper


class Mixer:
    """ Main Mixer class """

    _input_port = None
    _output_port = None
    _current_state = []
    _connected = False
    # noinspection PyTypeChecker
    _channels = {
        "lr": None,
        "inputs": [],
        "groups": [],
        "aux": [],
        "dcas": [],
        "mute_groups": [],
        "fx_returns": [],
        "fx_sends": [],
        "matrices": [],
    }

    def __init__(self, input_name="MIDI Control 1", output_name="MIDI Control 1", channel=0, debug=False):
        self._input = input_name
        self._output = output_name
        self._channel = channel
        self._debug = debug
        self._connect()
        self.init_channels()

    def _print_debug(self, msg):
        if self._debug:
            print("DEBUG [Mixer] => " + str(msg))

    def _connect(self):
        input_name = None
        output_name = None
        # Input
        self._print_debug("Connecting to input ...")
        self._print_debug(mido.get_input_names())
        for port in mido.get_input_names():
            if self._input in port:
                input_name = port
        # Output
        self._print_debug("Connecting to output ...")
        self._print_debug(mido.get_output_names())
        for port in mido.get_output_names():
            if self._output in port:
                output_name = port

        self._print_debug("Selected Input: " + str(input_name))
        self._print_debug("Selected Output: " + str(output_name))
        if input_name is None or output_name is None:
            return False
        # Connect
        try:
            self._input_port = mido.open_input(input_name)
            self._output_port = mido.open_output(output_name)
            self._print_debug("Connected to Input: '" + input_name + "' Output: '" + output_name + "'")
        except Exception as e:
            self._print_debug(e)
            return False
        self._connected = True
        return True

    @property
    def connected(self):
        """ Whether connected to the mixer (via MIDI) """
        return self._connected

    def __send_command(self, msb: int, lsb: int, data1: list[int], data2: list[int] = None):
        """ Send an NRPN command to the mixer """
        if not self.connected and not self._connect():
            return False
        temp = " ".join(map(str, data2)) if data2 else "[None]"
        self._print_debug("Sending command: MSB: " + str(msb) + " LSB: " + str(lsb) + " Data1: " + " ".join(
            map(str, data1)) + " Data2: " + temp)
        ch = 'B' + str(self._channel)
        # MSB & LSB
        msg_msb = Message.from_bytes([int(ch, 16), 99, msb])
        msg_lsb = Message.from_bytes([int(ch, 16), 98, msb])
        # Data
        msg_data1 = Message.from_bytes([int(ch, 16), *data1])
        # Send command
        self._output_port.send(msg_msb)
        self._output_port.send(msg_lsb)
        self._output_port.send(msg_data1)
        # If data2
        if data2 is not None:
            msg_data2 = Message.from_bytes([int(ch, 16), *data2])
            self._output_port.send(msg_data2)
        return True

    def set_param(self, msb_lsb: list[int], coarse: int, fine: int = None):
        """ Send a 'set' NRPN command with 'coarse' and 'fine' values """
        msb = msb_lsb[0]
        lsb = msb_lsb[1]
        # If only sending single parameter (e.g. mute)
        if fine is None:
            fine = coarse
            coarse = 0
        return self.__send_command(
            msb=msb,
            lsb=lsb,
            data1=[6, coarse],
            data2=[38, fine]
        )

    def get_param(self, msb_lsb: list[int]):
        """ Get a parameter from the mixer """
        if not self.connected and not self._connect():
            return None
        self.__send_command(
            msb=msb_lsb[0],
            lsb=msb_lsb[1],
            data1=[90, 127]
        )
        start = time.time_ns()
        msg = None
        while time.time_ns() - start < 5000 * 1000:
            if not self._input_port.closed:
                msg = self._input_port.poll()
        return msg if msg is not None else False

    def increment_param(self, msb, lsb):
        """ Increment a parameter """
        return self.__send_command(
            msb=msb,
            lsb=lsb,
            data1=[96, 0]
        )

    def decrement_param(self, msb, lsb):
        """ Decrement a parameter """
        return self.__send_command(
            msb=msb,
            lsb=lsb,
            data1=[97, 0]
        )

    """ Handle channels """

    def init_channels(self):
        """ Initialize all channels """
        # Avoid circular loop
        import sq_midi.channels as channels
        """ Initialize channels """
        # Inputs
        for i in range(channels.Input.NUMBER_OF_CHANNELS):
            self._channels["inputs"].append(channels.Input(self, i + 1))
        # Groups
        for i in range(channels.Group.NUMBER_OF_CHANNELS):
            self._channels["groups"].append(channels.Group(self, i + 1))
        # Auxes
        for i in range(channels.Aux.NUMBER_OF_CHANNELS):
            self._channels["aux"].append(channels.Aux(self, i + 1))
        # Matrices
        for i in range(channels.Matrix.NUMBER_OF_CHANNELS):
            self._channels["matrices"].append(channels.Matrix(self, i + 1))
        # LR
        # noinspection PyTypeChecker
        self._channels["lr"] = channels.LR(self)
        # FX Sends
        for i in range(channels.FXSend.NUMBER_OF_CHANNELS):
            self._channels["fx_sends"].append(channels.FXSend(self, i + 1))
        # FX Returns
        for i in range(channels.FXReturn.NUMBER_OF_CHANNELS):
            self._channels["fx_returns"].append(channels.FXReturn(self, i + 1))

    # Channels
    @property
    def inputs(self):
        """ List of input channels"""
        return ChannelWrapper(self._channels["inputs"])

    @property
    def groups(self):
        """ List of groups """
        return ChannelWrapper(self._channels["groups"])

    @property
    def lr(self):
        """ Main LR channel"""
        from sq_midi.channels import LR
        return LR(self)

    @property
    def aux(self):
        """ List of aux channels """
        return ChannelWrapper(self._channels["aux"])

    @property
    def matrices(self):
        """ List of matrices """
        return ChannelWrapper(self._channels["matrices"])

    @property
    def fxsends(self):
        """ List of FX Sends """
        return ChannelWrapper(self._channels["fx_sends"])

    @property
    def fxreturns(self):
        """ List of FX Returns """
        return ChannelWrapper(self._channels["fx_returns"])


if __name__ == "__main__":
    mixer = Mixer(
        input_name="MIDI Control 1",
        output_name="MIDI Control 1",
        debug=True
    )
    print(mixer.inputs[48].levels.lr)
    print(mixer.groups[1])
    #print(mixer.fxsends[1].panning)
    print(mixer.fxsends[1].assignments)



from typing import override

from sq_midi import Mixer
from sq_midi.channel import Channel
from sq_midi.data.loader import load_data

class LR(Channel):
    """ Main LR channel class """

    FILE_NAME = "lr"

    def __init__(self, mixer: Mixer):
        super().__init__(mixer)

    def __str__(self):
        return "LR object [Main LR]"

    def _load_info(self):
        return load_data(self.FILE_NAME)

if __name__ == "__main__":
    lr = LR(Mixer(debug=True))
    print(lr)
    lr.level = 1



from sq_midi.channel import Channel


class Group(Channel):
    """ Group Class """

    FILE_NAME = 'groups'
    CHANNEL_TYPE = 'Group Send'
    CHANNEL_PREFIX = 'Grp'
    NUMBER_OF_CHANNELS = 12

if __name__ == '__main__':
    from sq_midi import Mixer
    channel = Group(Mixer(debug=1), 1)
    print(channel.level)



from sq_midi.mixer import Mixer
from sq_midi.data.loader import load_data, decode


class Channel:
    """ Channel class
        Must be inherited by subclass to set FILE_NAME
    """

    FILE_NAME = None
    CHANNEL_TYPE = ""
    CHANNEL_PREFIX = ""
    NUMBER_OF_CHANNELS = -1
    CHANNEL_PARAMETERS = [
        "assignments",
        "levels",
        "panning"
    ]

    _level = 0
    _info = []
    _number = None
    __freeze = False

    def __init__(self, mixer: Mixer, number: int = None):
        self._mixer = mixer
        self._number = number
        if self.number is not None and self.number not in range(1, self.NUMBER_OF_CHANNELS + 1):
            raise ValueError(f"Channel number must be between 1 and {self.NUMBER_OF_CHANNELS}")
        self._info = self._load_info()
        self.__decode_info()
        self.__freeze = True

    def __str__(self):
        return f"{self.__class__.__name__} object [{self.CHANNEL_PREFIX}{self.number}]"

    def __decode_info(self):
        self._info = decode(self._info)
        try:
            self._info["levels"]["lr"] = [self._info["levels"]["lr"]]
            self._info["panning"]["lr"] = [self._info["panning"]["lr"]]
            self._info["assignments"]["lr"] = [self._info["assignments"]["lr"]]
        except KeyError:
            pass

    def _load_info(self):
        info = load_data(self.FILE_NAME)
        if self.number > len(info):
            raise ValueError("Channel number out of range")
        else:
            return info[self.number-1]

    @property
    def info(self):
        """ Channel commands & info """
        return self._info

    @property
    def number(self):
        """ Channel number"""
        return self._number

    @property
    def name(self):
        return self.info['name']

    """MUTE"""
    @property
    def mute(self):
        """ Get channel mute """
        return self._mixer.get_param(self.info["mute"])

    @mute.setter
    def mute(self, value: bool):
        """ Set channel mute """
        if type(value) is not bool and value not in [0, 1]:
            raise ValueError("Channel mute must be a boolean")
        self._mixer.set_param(self.info["mute"], int(value))

    def unmute(self):
        self.mute = False

    """LEVEL"""
    @property
    def level(self):
        """ Get current channel master level """
        return self.levels.lr

    @level.setter
    def level(self, value):
        """ Set channel master level """
        self.levels.lr = value

    """PANNING"""
    @property
    def pan(self):
        """ Get current channel master panning """
        return self.panning.lr

    @pan.setter
    def pan(self, value):
        """ Set channel master panning """
        self.panning.lr = value

    def center(self):
        self.panning.lr = 0

    def __getattr__(self, item):
        """ Handle other parameters as defined in JSON
            e.g. assignments / panning
            e.g. [channel].levels
                 [channel].panning
        """
        if item not in self.CHANNEL_PARAMETERS:
            raise AttributeError(f"{self.CHANNEL_TYPE} channel does not have parameter {item}"
                                 f"\nAvailable parameters: {", ".join(self.CHANNEL_PARAMETERS)}")
        t = None # Value data type
        r = None # Range of values
        m = lambda val: val # Map value to correct range
        match item:
            case "levels":
                t = int
                r = range(0, 110+1)
                # Map [0, 100] to [0, 12544] and [100, 110] to [12544, 16320]
                # according to MIDI datasheet [-inf, 0dB] and [0dB, +10dB]
                m = lambda val: int(
                        125.44 * val
                        if 0 <= val <= 100 else
                        (377.6 * (val - 100) + 12544 )
                    )
            case "assignments":
                t = bool
                r = [True, False]
            case "panning":
                t = int
                r = range(-100, 100+1)
                # Map [-100, 100] to [0, 16383]
                m = lambda val: int(((val + 100) / 200) * 16383)
        if t is not None:
            # Avoid circular dependency
            from .parameter_wrapper import ParameterWrapper
            # Use parameter wrapper class
            return ParameterWrapper(
                mixer = self._mixer,
                channel = self,
                parameter_category=item,
                parameter_type = t,
                parameter_range = r,
                parameter_map = m
            )
        else:
            # If parameter name does not exist, use normal attribute error message
            return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if (self.__freeze or key in self.CHANNEL_PARAMETERS) and key not in ["level", "mute", "pan"]:
            raise TypeError("Cannot edit instance of Channel"
                            "\n[Attributes] \nInfo: name, number\nControl: level, mute, pan\nRouting: levels, assignments, panning"
                            "\n[Methods]: \nunmute(), center()")
        object.__setattr__(self, key, value)


if __name__ == "__main__":
    _mixer = Mixer(
            input_name="MIDI Control 1",
            output_name="MIDI Control 1",
            debug = True
        )

    print(_mixer.inputs[1].mute)
    _mixer.inputs[1].levels = 100
    print(_mixer.inputs[1].levels)
    print(_mixer.inputs.ip133)


from sq_midi import Channel


class DCA(Channel):
    """ DCA channel class """

    CHANNEL_TYPE = CHANNEL_PREFIX = "DCA"
    NUMBER_OF_CHANNELS = 8

    def _load_info(self):
        return {
            "name": self.CHANNEL_PREFIX + str(self.number),
            "mute": ["02", "0" + str(self.number-1)], # [02 00] -> [02 07]
            "levels": {
                "lr": ["4F", "2" + str(self.number-1)], # [4F 20] -> [4F 27]
            }
        }

if __name__ == '__main__':
    from sq_midi import Mixer
    channel = DCA(Mixer(debug=1), 8)
    #print(channel.level)
    channel.mute = 1
    #print(channel.mute)
    #channel.mute = False
    #print(channel.info)



from sq_midi import Mixer
from sq_midi.channel import Channel

class Input(Channel):
    """ Input Channel Class """

    FILE_NAME = 'inputs'
    CHANNEL_TYPE = 'Input'
    CHANNEL_PREFIX = 'Ip'
    NUMBER_OF_CHANNELS = 48

if __name__ == '__main__':
    channel = Input(Mixer(debug=1), 1)
    print(channel.level)
    channel.panning.groups[0] =1


from sq_midi.channel import Channel


class Matrix(Channel):
    """ Matrix class """

    FILE_NAME = "matrices"
    CHANNEL_TYPE = "Matrices"
    CHANNEL_PREFIX = "Mtx"
    NUMBER_OF_CHANNELS = 3

    def _load_info(self):
        return {
            "name": self.CHANNEL_PREFIX + str(self.number),
            "mute": ["00", "5" + str(self.number+4)], # [00 55] -> [00 57]
            "levels": {
                "lr": ["4F", "1" + str(self.number)], # [4F 11] -> [4F 13]
            },
            "panning": {
                "lr": ["5F", "1" + str(self.number)], # [5F 11] -> [5F 13]
            }
        }

if __name__ == "__main__":
    from sq_midi import Mixer
    c = Matrix(Mixer(), 1)
    print(c.info)